@@ -161,50 +161,64 @@ pyi_path_join(char *result, const char *path1, const char *path2)
     return result;
 }
 
-
-#if !defined(_WIN32) && !defined(__APPLE__)
 /*
- * Return full path to a file's directory, but keeps the basename.
- * This is required to pass the correct basename to execvp().
+ * Fully resolve the given path; canonicalize it and resolve symbolic
+ * links. It is implicitly assumed that the output buffer is large
+ * enough to accept up to PATH_MAX characters (including the terminating
+ * NULL character).
  */
 int
-pyi_path_fullpath_keep_basename(char *abs, const char *rel)
-{
-    char dirname[PATH_MAX];
-    char full_dirname[PATH_MAX];
-    char basename[PATH_MAX];
-    pyi_path_basename(basename, rel);
-    pyi_path_dirname(dirname, rel);
-    if (realpath(dirname, full_dirname) == NULL) {
-        return false;
-    }
-    return (pyi_path_join(abs, full_dirname, basename) != NULL);
-}
-#endif
-
-/*
- * Return full path to a file. Wraps platform specific function.
- */
-int
-pyi_path_fullpath(char *abs, size_t abs_size, const char *rel)
+pyi_path_resolve(const char *path, char *resolved_path)
 {
 #ifdef _WIN32
-    wchar_t wrel[PATH_MAX + 1];
-    wchar_t *wabs = NULL;
+    wchar_t wpath[PATH_MAX + 1];
+    wchar_t wresolved_path[PATH_MAX + 1];
 
-    pyi_win32_utils_from_utf8(wrel, rel, PATH_MAX);
+    HANDLE handle;
+    DWORD ret;
 
-    wabs = _wfullpath(NULL, wrel, PATH_MAX);
-    if (wabs == NULL) {
+    int offset = 0;
+
+    pyi_win32_utils_from_utf8(wpath, path, PATH_MAX);
+
+    /* Open file/directory handle */
+    handle = CreateFileW(
+        wpath, /* lpFileName */
+        0, /* dwDesiredAccess */
+        FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, /* dwShareMode */
+        NULL, /* lpSecurityAttributes */
+        OPEN_EXISTING, /* dwCreationDisposition */
+        FILE_ATTRIBUTE_NORMAL, /* dwFlagsAndAttributes*/
+        NULL /* hTemplateFile */
+    );
+    if (handle == INVALID_HANDLE_VALUE) {
         return 0;
     }
 
-    char *ret = pyi_win32_utils_to_utf8(abs, wabs, abs_size);
-    free(wabs);
+    /* Fully resolve the path */
+    ret = GetFinalPathNameByHandleW(
+        handle,  /* hFile */
+        wresolved_path, /* lpszFilePath */
+        PATH_MAX, /* cchFilePath */
+        FILE_NAME_NORMALIZED /* dwFlags */
+    );
 
-    return ret != NULL;
+    CloseHandle(handle);
+
+    if (ret == 0 || ret >= PATH_MAX) {
+        /* Failure or insufficient buffer size */
+        return  0;
+    }
+
+    /* Remove the extended path indicator, to avoid potential issues due
+     * to its appearance in `sys.executable`, `sys._MEIPASS`, etc. */
+    if (ret >= 4 && wcsncmp(L"\\\\?\\", wresolved_path, 4) == 0) {
+        offset = 4;
+    }
+
+    return pyi_win32_utils_to_utf8(resolved_path, wresolved_path + offset, PATH_MAX) != NULL;
 #else
-    return realpath(rel, abs) != NULL;
+    return realpath(path, resolved_path) != NULL;
 #endif
 }
 
@@ -315,19 +329,19 @@ pyi_path_executable(char *execfile, const char *appname)
     }
 
 #elif __APPLE__
-    char buffer[PATH_MAX];
-    uint32_t length = sizeof(buffer);
+    char program_path[PATH_MAX];
+    uint32_t length = sizeof(program_path);
 
     /* Mac OS X has special function to obtain path to executable.
      * This may return a symlink.
      */
-    if (_NSGetExecutablePath(buffer, &length) != 0) {
-        FATALERROR("System error - unable to load!\n");
+    if (_NSGetExecutablePath(program_path, &length) != 0) {
+        FATALERROR("Failed to obtain executable path via _NSGetExecutablePath!\n");
         return false;
     }
 
-    if (pyi_path_fullpath(execfile, PATH_MAX, buffer) == false) {
-        VS("LOADER: Cannot get fullpath for %s\n", execfile);
+    if (pyi_path_resolve(program_path, execfile) == false) {
+        VS("LOADER: failed to resolve full path for %s\n", program_path);
         return false;
     }
 
@@ -360,53 +374,65 @@ pyi_path_executable(char *execfile, const char *appname)
 #endif
 
     if (name_len == -1) {
+        /* We failed to resolve the executable file via /proc (or we were
+         * launched via ld.so dynamic loader). Try to manually resolve the
+         * program path/name given via argv[0]. */
         if (strchr(appname, PYI_SEP)) {
-            /* Absolute or relative path: Canonicalize directory path,
-             * but keep original basename.
-             */
-            if (pyi_path_fullpath_keep_basename(execfile, appname) == false) {
-                VS("LOADER: Cannot get fullpath for %s\n", execfile);
+            /* Absolute or relative path was given. Canonicalize it, and
+             * resolve symbolic links. */
+            VS("LOADER: resolving program path %s...\n", appname);
+            if (pyi_path_resolve(appname, execfile) == false) {
+                VS("LOADER: failed to resolve full path for %s\n", appname);
                 return false;
             }
-        }
-        else {
-            /* No absolute or relative path, just program name: search $PATH.
-             */
-            char buffer[PATH_MAX];
-            if (! pyi_search_path(buffer, appname)) {
-                /* Searching $PATH failed, user is crazy. */
-                VS("LOADER: Searching $PATH failed for %s\n", appname);
-                if (snprintf(buffer, PATH_MAX, "%s", appname) >= PATH_MAX) {
-                    VS("LOADER: Full path to application exceeds PATH_MAX: %s\n", appname);
+        } else {
+            /* No path, just program name. Search $PATH for executable
+             * with matching name. */
+            char program_path[PATH_MAX];
+            if (pyi_search_path(program_path, appname)) {
+                /* Program found in $PATH; resolve full path */
+                VS("LOADER: program %s found in PATH: %s. Resolving full path...\n", appname, program_path);
+                if (pyi_path_resolve(program_path, execfile) == false) {
+                    VS("LOADER: failed to resolve full path for %s\n", program_path);
+                    return false;
+                }
+            } else {
+                /* Searching $PATH failed; try resolving the name as-is,
+                 * and hope for the best. NOTE: can we even reach this
+                 * part? How was the executable even launched in such
+                 * case? */
+                VS("LOADER: could not find %s in $PATH! Attempting to resolve as-is...\n", appname);
+                if (pyi_path_resolve(appname, execfile) == false) {
+                    VS("LOADER: failed to resolve full path for %s\n", appname);
                     return false;
                 }
             }
-            if (pyi_path_fullpath_keep_basename(execfile, buffer) == false) {
-                VS("LOADER: Cannot get fullpath for %s\n", execfile);
-                return false;
-            }
-        }
-    }
-
-    /* Check if execfile is a symbolic link. Usually, the /proc entry resolution
-     * ensures that it is a regular file path. However, in some cases, /proc entry
-     * resolution is unavailable (e.g., launching via ld.so dynamic loader), and
-     * we need to deal with symlinks ourselves... */
-    if (pyi_path_is_symlink(execfile)) {
-        /* Create a copy of original name; we need this to resolve relative symbolic
-         * link (as well as for the error message). */
-        char orig_execfile[PATH_MAX];
-        if (snprintf(orig_execfile, PATH_MAX, "%s", execfile) >= PATH_MAX) {
-            return false;
-        }
-        /* Fully resolve the path */
-        if (pyi_path_fullpath(execfile, PATH_MAX, orig_execfile) == false) {
-            VS("LOADER: Cannot resolve executable symbolic link %s\n", orig_execfile);
-            return false;
         }
     }
 
 #endif /* ifdef _WIN32 */
+
+    /* Check if execfile is a symbolic link. The macOS and POSIX codepaths should
+     * already take care of resolving symbolic links, while Windows codepath does
+     * not. Keep this as a common step nevertheless, just in case. */
+    if (pyi_path_is_symlink(execfile)) {
+        char orig_execfile[PATH_MAX];
+
+        VS("LOADER: executable %s is a symbolic link - resolving...\n", execfile);
+
+        /* Create a copy of original name; we need this to resolve relative symbolic
+         * link (as well as for the error message). */
+        if (snprintf(orig_execfile, PATH_MAX, "%s", execfile) >= PATH_MAX) {
+            return false;
+        }
+
+        /* Fully resolve the path */
+        if (pyi_path_resolve(orig_execfile, execfile) == false) {
+            VS("LOADER: failed to resolve executable symbolic link %s\n", orig_execfile);
+            return false;
+        }
+    }
+
     VS("LOADER: executable is %s\n", execfile);
     return true;
 }
